/* Windows GDI+ related functions */
#include <windows.h>
#include <gdiplus.h>
#include <wchar.h>

#include <sp/spBaseLib.h>
#include <sp/spComponentLib.h>
#include <sp/spCanvasP.h>
#include <sp/spGraphicsP.h>
#include <sp/spDrawP.h>

#include <sp/spDIBitmapP.h>

static spBool sp_wingp_initialized = SP_FALSE;

#undef SP_USE_SP_FUNCTION_FOR_BMP_WITH_ALPHA

#undef SP_INVALIDATE_READ_SUPPORT_FOR_DEBUG
//#define SP_INVALIDATE_READ_SUPPORT_FOR_DEBUG

typedef struct _spDIBitmapInstanceArch {
    int type_index;
    WCHAR* wfilename;
    char mode[8];
    CLSID clsid;
    Gdiplus::Bitmap* bitmap;

    spBool bmp_with_alpha;
#if defined(SP_USE_SP_FUNCTION_FOR_BMP_WITH_ALPHA)
    FILE *fp;
    DWORD bfSize, bfOffBits;
    BITMAPV5HEADER header;
#endif
} *spDIBitmapInstanceArch;

spBool spInitWinGP(void)
{
    ULONG_PTR gpToken = 0;
    Gdiplus::GdiplusStartupInput gpStartupInput;
    
    if (sp_wingp_initialized == SP_TRUE) {
	return SP_TRUE;
    }
    
    if (Gdiplus::GdiplusStartup(&gpToken, &gpStartupInput, NULL) != Gdiplus::Ok) {
	return SP_FALSE;
    }
    
    spAddExitCallback(spShutdownWinGP, (void *)gpToken);

    sp_wingp_initialized = SP_TRUE;
    
    return SP_TRUE;
}

void spShutdownWinGP(void *data)
{
    ULONG_PTR gpToken = (ULONG_PTR)data;

    Gdiplus::GdiplusShutdown(gpToken);
    sp_wingp_initialized = SP_FALSE;
    
    return;
}

Gdiplus::PixelFormat spGetPixelFormatWinGP(spDIBitmapInfoInternal *info)
{
    Gdiplus::PixelFormat gppformat = 0;

    if (((info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)
	 && !(info->pixel_format & SP_DI_PIXEL_FORMAT_LAST_ALPHA_MASK))
	|| !(info->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK)) {
        spDebug(80, "spGetPixelFormatWinGP", "not supported format = %lx\n", info->pixel_format);
	return 0;
    }
	
    if (info->bit_size == 16) {
	if (info->pixel_stride == 4) {
	    if (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) {
		if (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) {
		    if (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK) {
			gppformat = PixelFormat64bppPARGB;
		    } else {
			gppformat = PixelFormat64bppARGB;
		    }
		}
	    }
	} else if (info->pixel_stride == 3) {
	    if (!(info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)) {
		gppformat = PixelFormat48bppRGB;
	    }
	}
    } else if (info->bit_size == 8) {
	if (info->pixel_stride == 4) {
	    if (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) {
                spDebug(80, "spGetPixelFormatWinGP", "with alpha\n");
		if (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK) {
		    gppformat = PixelFormat32bppPARGB;
		} else {
		    gppformat = PixelFormat32bppARGB;
		}
	    } else {
                spDebug(80, "spGetPixelFormatWinGP", "without alpha\n");
		gppformat = PixelFormat32bppRGB;
	    }
	} else if (info->pixel_stride == 3) {
	    if (!(info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)) {
		gppformat = PixelFormat24bppRGB;
	    }
	}
    }

    spDebug(80, "spGetPixelFormatWinGP", "gpformat = %x\n", gppformat);
    
    return gppformat;
}

spBool spCopyDIBitmapToImageWinGP(spDIBitmapInfoInternal *info, void *bitmap, spComponent dest, int src_x, int src_y,
				  int width, int height, int dest_x, int dest_y)
{
    INT stride;
    BYTE *ptr;
    Gdiplus::PixelFormat format = spGetPixelFormatWinGP(info);
    spDebug(80, "spCopyDIBitmapToImageWinGP", "format = %x\n", format);
    if (format <= 0) {
	return SP_FALSE;
    }

    if (info->upside_down) {
        stride = -info->line_stride;
        ptr = (BYTE *)info->data + info->line_stride * (info->height - 1);
    } else {
        stride = info->line_stride;
        ptr = (BYTE *)info->data;
    }
    
    Gdiplus::Graphics graphics(SpPrimitiveArch(dest).memdc);
    Gdiplus::Bitmap gpbitmap(info->width, info->height, stride, format, ptr);

    if (graphics.DrawImage(&gpbitmap, dest_x, dest_y, src_x, src_y, width, height, Gdiplus::UnitPixel) == Gdiplus::Ok) {
	return SP_TRUE;
    } else {
	spDebug(10, "spCopyDIBitmapToImageWinGP", "DrawImage failed\n");
	return SP_FALSE;
    }
}

spBool spStretchDIBitmapToImageWinGP(spDIBitmapInfoInternal *info, void *bitmap, spComponent dest,
				     int src_x, int src_y, int src_width, int src_height,
				     int dest_x, int dest_y, int dest_width, int dest_height)
{
    INT stride;
    BYTE *ptr;
    Gdiplus::PixelFormat format = spGetPixelFormatWinGP(info);
    spDebug(80, "spStretchDIBitmapToImageWinGP", "format = %x\n", format);
    if (format <= 0) {
	return SP_FALSE;
    }
    
    if (info->upside_down) {
        stride = -info->line_stride;
        ptr = (BYTE *)info->data + info->line_stride * (info->height - 1);
    } else {
        stride = info->line_stride;
        ptr = (BYTE *)info->data;
    }
    
    Gdiplus::Graphics graphics(SpPrimitiveArch(dest).memdc);
    Gdiplus::Bitmap gpbitmap(info->width, info->height, stride, format, ptr);

    Gdiplus::Rect destRect(dest_x, dest_y, dest_width, dest_height);

    if (graphics.DrawImage(&gpbitmap, destRect, src_x, src_y, src_width, src_height,
			   Gdiplus::UnitPixel, NULL, NULL, NULL) == Gdiplus::Ok) {
	return SP_TRUE;
    } else {
	spDebug(10, "spStretchDIBitmapToImageWinGP", "DrawImage failed\n");
	return SP_FALSE;
    }
}

void *spCopyImageToDIBitmapWinGP(spComponent src, spDIBitmapInfoInternal *info, void *bitmap, int src_x, int src_y,
				int width, int height, int dest_x, int dest_y)
{
    INT stride;
    BYTE *ptr;
    
    if (SpPrimitiveArch(src).hbitmap == NULL) {
        spDebug(10, "spCopyDIBitmapToImageWinGP", "no hbitmap\n");
        return NULL;
    }
    
    Gdiplus::PixelFormat format = spGetPixelFormatWinGP(info);
    spDebug(80, "spCopyDIBitmapToImageWinGP", "format = %x\n", format);
    if (format <= 0) {
	return NULL;
    }
    
    if (info->upside_down) {
        stride = -info->line_stride;
        ptr = (BYTE *)info->data + info->line_stride * (info->height - 1);
    } else {
        stride = info->line_stride;
        ptr = (BYTE *)info->data;
    }
    
    Gdiplus::Bitmap srcBitmap(SpPrimitiveArch(src).hbitmap, NULL);
    Gdiplus::Bitmap destBitmap(info->width, info->height, stride, format, ptr);

    Gdiplus::Graphics graphics(&destBitmap);
    
    if (graphics.DrawImage(&srcBitmap, dest_x, dest_y, src_x, src_y, width, height, Gdiplus::UnitPixel) == Gdiplus::Ok) {
	spDebug(80, "spCopyImageToDIBitmapWinGP", "DrawImage done\n");
	return bitmap;
    } else {
	spDebug(10, "spCopyImageToDIBitmapWinGP", "DrawImage failed\n");
	return NULL;
    }
}

static WCHAR *xspGetWideStringWin(const char *str)
{
    int nchar;
    WCHAR *wstr;
    
    nchar = MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, (LPSTR)str, -1, NULL, 0);
    spDebug(80, "xspGetWideStringWin", "str = %s, nchar = %d\n", str, nchar);
    
    if ((wstr = xalloc(nchar + 1, WCHAR)) != NULL) {
	if (MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, (LPSTR)str, -1, wstr, nchar) == 0) {
	    xfree(wstr);
	    wstr = NULL;
	}
	/*fwprintf(stderr, L"wstr = %s\n", wstr);*/
    }

    return wstr;
}

void *spInitDIBitmapInstanceArch(const char *name)
{
    spDIBitmapInstanceArch dibinstance = NULL;

    if (spInitWinGP() == SP_FALSE) {
	spDebug(80, "spInitDIBitmapInstanceArch", "spInitWinGP failed\n");
	return NULL;
    }

    dibinstance = xalloc(1, struct _spDIBitmapInstanceArch);
    memset(dibinstance, 0, sizeof(struct _spDIBitmapInstanceArch));
    dibinstance->type_index = -1;
    
    spDebug(80, "spInitDIBitmapInstanceArch", "done\n");
    
    return dibinstance;
}

spBool spFreeDIBitmapInstanceArch(void *instance)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;

    xfree(dibinstance);
    
    spDebug(80, "spFreeDIBitmapInstanceArch", "done\n");
    
    return SP_TRUE;
}

int spGetDIBitmapPriorityArch(void *instance)
{
    return SP_PLUGIN_PRIORITY_MIDDLE;
}

unsigned long spGetDIBitmapCapsArch(void *instance)
{
    unsigned long caps;
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    
    caps = SP_DI_BITMAP_CAPS_READ | SP_DI_BITMAP_CAPS_WRITE
	| SP_DI_BITMAP_CAPS_DIRECT_ACCESS | SP_DI_BITMAP_CAPS_PIXEL_STRIDE_3 | SP_DI_BITMAP_CAPS_PIXEL_STRIDE_4
	| SP_DI_BITMAP_CAPS_COPY_IMAGE | SP_DI_BITMAP_CAPS_COPY_IMAGE_STRETCHABLE
	| SP_DI_BITMAP_CAPS_UPSIDE_UP | SP_DI_BITMAP_CAPS_UPSIDE_DOWN 
	| SP_DI_BITMAP_CAPS_INVERSE_FORMAT | SP_DI_BITMAP_CAPS_LAST_ALPHA;

    if (dibinstance->type_index == 0 || dibinstance->type_index == 4) {
	caps |= SP_DI_BITMAP_CAPS_ALPHA;
    }

    return caps;
}

static const char *sp_dibitmap_file_type_list_arch[] = {
    "png",
    "bmp",
    "jpeg",
    "gif",
    "tiff",
    NULL,
};

static const char *sp_dibitmap_file_desc_list_arch[] = {
    "PNG (GDI+)",
    "BMP (GDI+)",
    "JPEG (GDI+)",
    "GIF (GDI+)",
    "TIFF (GDI+)",
    NULL,
};

static const char *sp_dibitmap_file_filter_list_arch[] = {
    "*.png",
    "*.bmp",
    "*.jpg",
    "*.gif",
    "*.tif",
    NULL,
};

const char **spGetDIBitmapFileTypeListArch(void *instance)
{
    return sp_dibitmap_file_type_list_arch;
}

const char **spGetDIBitmapFileDescListArch(void *instance)
{
    return sp_dibitmap_file_desc_list_arch;
}

const char **spGetDIBitmapFileFilterListArch(void *instance)
{
    return sp_dibitmap_file_filter_list_arch;
}

spBool spIsDIBitmapSupportedArch(void *instance, const char *filename)
{
    spBool supported = SP_FALSE;
    
    spDebug(80, "spIsDIBitmapSupportedArch", "filename = %s\n", filename);
    
    if (spGetBaseName(filename) == filename && filename[0] == '.') {
	if (strcaseeq(filename, ".png")) {
	    supported = SP_TRUE;
	} else if (strcaseeq(filename, ".bmp")) {
	    supported = SP_TRUE;
	} else if (strcaseeq(filename, ".gif")) {
	    supported = SP_TRUE;
	} else if (strcaseeq(filename, ".jpg") || strcaseeq(filename, ".jpeg")) {
	    supported = SP_TRUE;
	} else if (strcaseeq(filename, ".tif") || strcaseeq(filename, ".tiff")) {
	    supported = SP_TRUE;
	}
	spDebug(80, "spIsDIBitmapSupportedArch", "write mode: supported = %d\n", supported);
    } else {
#if !defined(SP_INVALIDATE_READ_SUPPORT_FOR_DEBUG)
	WCHAR* wfilename = xspGetWideStringWin(filename);

	if (wfilename != NULL) {
	    Gdiplus::Status status;
	    Gdiplus::Bitmap bitmap(wfilename, FALSE);
	    
	    if ((status = bitmap.GetLastStatus()) == Gdiplus::Ok) {
		supported = SP_TRUE;
	    }
	    spDebug(80, "spIsDIBitmapSupportedArch", "after bitmap.GetLastStatus(): status = %d, supported = %d\n",
		    (int)status, supported);
	    
	    xfree(wfilename);
	}
#endif
    }

    return supported;
}

spBool spSetDIBitmapFileTypeArch(void *instance, int index)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;

    dibinstance->type_index = index;
    
    return SP_TRUE;
}

int spGetDIBitmapFileTypeArch(void *instance)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    
    return dibinstance->type_index;
}

const char *spGetDIBitmapBestSuffixArch(void *instance)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    
    if (dibinstance->type_index == 0) {
	return ".png";
    } else if (dibinstance->type_index == 1) {
	return ".bmp";
    } else if (dibinstance->type_index == 2) {
	return ".jpg";
    } else if (dibinstance->type_index == 3) {
	return ".gif";
    } else if (dibinstance->type_index == 4) {
	return ".tif";
    } else {
	return NULL;
    }
}

static const WCHAR *getFileTypeMimeString(int type_index)
{
    const WCHAR* mimeString = NULL;

    if (type_index < 0) return NULL;
    
    if (type_index == 0) {
	mimeString = L"image/png";
    } else if (type_index == 1) {
	mimeString = L"image/bmp";
    } else if (type_index == 2) {
	mimeString = L"image/jpeg";
    } else if (type_index == 3) {
	mimeString = L"image/gif";
    } else if (type_index == 4) {
	mimeString = L"image/tiff";
    }

    return mimeString;
}

static const WCHAR *getMimeStringFromFilename(const char *filename)
{
    char* suffix;
    const WCHAR* mimeString = NULL;
	
    if ((suffix = spGetSuffix(filename)) == NULL) {
	return NULL;
    }
    
    if (strcaseeq(suffix, ".png")) {
	mimeString = L"image/png";
    } else if (strcaseeq(suffix, ".bmp")) {
	mimeString = L"image/bmp";
    } else if (strcaseeq(suffix, ".jpg") || strcaseeq(suffix, ".jpeg")) {
	mimeString = L"image/jpeg";
    } else if (strcaseeq(suffix, ".gif")) {
	mimeString = L"image/gif";
    } else if (strcaseeq(suffix, ".tif") || strcaseeq(suffix, ".tiff")) {
	mimeString = L"image/tiff";
    }

    return mimeString;
}

static int getFileTypeIndexFromBitmap(Gdiplus::Bitmap* gpbitmap)
{
    int type_index = -1;

    GUID guid;
    if (gpbitmap->GetRawFormat(&guid) == Gdiplus::Ok) {
	if(guid == Gdiplus::ImageFormatPNG) {
	    type_index = 0;
	} else if(guid == Gdiplus::ImageFormatBMP) {
	    type_index = 1;
	} else if(guid == Gdiplus::ImageFormatJPEG) {
	    type_index = 2;
	} else if(guid == Gdiplus::ImageFormatGIF) {
	    type_index = 3;
	} else if(guid == Gdiplus::ImageFormatTIFF) {
	    type_index = 4;
	}
    }

    return type_index;
}

static int getEncoderClsid(const WCHAR* format, CLSID* pClsid)
{
    UINT num = 0;
    UINT size = 0;
    Gdiplus::ImageCodecInfo* pImageCodecInfo = NULL;

    Gdiplus::GetImageEncodersSize(&num, &size);
    if (size == 0)
	return -1;

    pImageCodecInfo = (Gdiplus::ImageCodecInfo*)(malloc(size));
    if (pImageCodecInfo == NULL)
	return -1;

    Gdiplus::GetImageEncoders(num, size, pImageCodecInfo);

    for (UINT j = 0; j < num; ++j) {
	if (wcscmp(pImageCodecInfo[j].MimeType, format) == 0) {
	    *pClsid = pImageCodecInfo[j].Clsid;
	    free(pImageCodecInfo);
	    return j;
	}    
    }

    free(pImageCodecInfo);
    return -1;
}

spBool spOpenDIBitmapFileArch(void *instance, const char *filename, const char *mode)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;

    spDebug(80, "spOpenDIBitmapFileArch", "filename = %s, mode = %s, type_index = %d\n",
	    filename, mode, dibinstance->type_index);

    if (mode[0] == 'w') {
	const WCHAR* mimeString;
	
	if ((mimeString = getFileTypeMimeString(dibinstance->type_index)) == NULL) {
	    if ((mimeString = getMimeStringFromFilename(filename)) == NULL) {
		return SP_FALSE;
	    }
	}
	
	CLSID clsid;
	if (getEncoderClsid(mimeString, &clsid) < 0) {
	    return SP_FALSE;
	}
	
	WCHAR* wfilename = xspGetWideStringWin(filename);

	if (wfilename == NULL) {
	    return SP_FALSE;
	}

	dibinstance->wfilename = wfilename;
	dibinstance->clsid = clsid;
	dibinstance->bitmap = NULL;
    } else {
	dibinstance->wfilename = NULL;
        
#if !defined(SP_INVALIDATE_READ_SUPPORT_FOR_DEBUG)
        dibinstance->bmp_with_alpha = SP_FALSE;
#if defined(SP_USE_SP_FUNCTION_FOR_BMP_WITH_ALPHA)
        dibinstance->fp = NULL;
        dibinstance->bfSize = dibinstance->bfOffBits = 0;
    
#if 1
        {
            if ((dibinstance->fp = spOpenFile(filename, "rb")) != NULL) {
                if (spReadBmpFileHeader(&dibinstance->bfSize, &dibinstance->bfOffBits, dibinstance->fp) > 0) {
                    if (spReadBmpV5Header(&dibinstance->header, dibinstance->bfOffBits, dibinstance->fp) > 0) {
                        if (dibinstance->bfSize >= 56 && dibinstance->header.bV5BitCount == 32
                            && dibinstance->header.bV5Compression == 3 && dibinstance->header.bV5AlphaMask == 0xff000000) {
                            dibinstance->bmp_with_alpha = SP_TRUE;
                        }
                    }
                }
                
                if (dibinstance->bmp_with_alpha == SP_FALSE) {
                    spCloseFile(dibinstance->fp);
                    dibinstance->fp = NULL;
                }
            }
        }
#endif
#endif

        if (dibinstance->bmp_with_alpha == SP_FALSE) {
            WCHAR* wfilename = xspGetWideStringWin(filename);

            if (wfilename == NULL) {
                return SP_FALSE;
            }

            Gdiplus::Bitmap* gpbitmap = Gdiplus::Bitmap::FromFile(wfilename);

            if (gpbitmap == NULL || gpbitmap->GetLastStatus() != Gdiplus::Ok) {
                spDebug(10, "spOpenDIBitmapFileArch", "Bitmap::FromFile failed\n");
                if (gpbitmap != NULL) {
                    delete gpbitmap;
                }
                xfree(wfilename);
                return SP_FALSE;
            }

            dibinstance->type_index = getFileTypeIndexFromBitmap(gpbitmap);
	
            dibinstance->wfilename = wfilename;
            dibinstance->bitmap = gpbitmap;
        }
#else
	dibinstance->wfilename = NULL;
	dibinstance->bitmap = NULL;
	return SP_FALSE;
#endif
    }

    spStrCopy(dibinstance->mode, sizeof(dibinstance->mode), mode);
    
    spDebug(80, "spOpenDIBitmapFileArch", "done\n");
    
    return SP_TRUE;
}

spBool spCloseDIBitmapFileArch(void *instance)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    
    if (dibinstance->bmp_with_alpha) {
#if defined(SP_USE_SP_FUNCTION_FOR_BMP_WITH_ALPHA)
        if (dibinstance->fp != NULL) {
            spCloseFile(dibinstance->fp);
            dibinstance->fp = NULL;
        }
#endif
    } else {
        if (dibinstance->wfilename != NULL) {
            xfree(dibinstance->wfilename);
            dibinstance->wfilename = NULL;
        }
    
        if (dibinstance->bitmap != NULL) {
            delete dibinstance->bitmap;
            dibinstance->bitmap = NULL;
        }
    }

    dibinstance->bmp_with_alpha = SP_FALSE;
    dibinstance->type_index = -1;
    dibinstance->mode[0] = NUL;

    return SP_TRUE;
}

spBool spGetDIBitmapInfoArch(void *instance, spDIBitmapInfoInternal *info, unsigned long *strict_mask)
{
    spBool flag = SP_FALSE;
    Gdiplus::Bitmap* gpbitmap;
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;

    spDebug(80, "spGetDIBitmapInfoArch", "in: strict mask = %lx\n", *strict_mask);

    if (dibinstance->bmp_with_alpha) {
#if defined(SP_USE_SP_FUNCTION_FOR_BMP_WITH_ALPHA)
        if (dibinstance->fp == NULL) return SP_FALSE;
        
        return spGetDIBInfoFromBmpV5Header(&dibinstance->header, info, strict_mask, dibinstance->mode[0] == 'w' ? SP_TRUE : SP_FALSE);
#endif
    }
    
    if (((*strict_mask & SP_DI_BITMAP_STRICT_GRAYSCALE_STATE)
	 && (info->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_YUV_STATE)
	    && (info->pixel_format & SP_DI_PIXEL_FORMAT_YUV_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_ORDER)
	    && !(info->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_BIT_SIZE)
	    && info->bit_size != 8)) {
	spDebug(80, "spGetDIBitmapInfoArch", "unsupported strict mask: %lx\n", *strict_mask);
	return SP_FALSE;
    }
	
    if ((gpbitmap = dibinstance->bitmap) == NULL) return SP_FALSE;
    
    Gdiplus::PixelFormat format = gpbitmap->GetPixelFormat();
    spDIPixelFormat pixel_format = 0;
    int pixel_stride = 0;

    if (format == PixelFormat24bppRGB) {
	if (((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
	     && info->pixel_stride != 3)
	    || ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_STATE)
		&& (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK))) {
	    return SP_FALSE;
	}
	pixel_stride = 3;
	pixel_format = SP_DI_PIXEL_FORMAT_BGR;
    } else if (format == PixelFormat32bppRGB) {
	if (((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
	     && info->pixel_stride != 4)
	    || ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_STATE)
		&& (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK))) {
	    return SP_FALSE;
	}
	pixel_stride = 4;
	pixel_format = SP_DI_PIXEL_FORMAT_BGRN;
    } else if (format == PixelFormat32bppARGB) {
	if (((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
	     && info->pixel_stride != 4)
	    || ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_STATE)
		&& !(info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK))
	    || ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_PREMULTIPLIED_STATE)
		&& (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK))) {
	    return SP_FALSE;
	}
	pixel_stride = 4;
	pixel_format = SP_DI_PIXEL_FORMAT_BGRA;
    } else if (format == PixelFormat32bppPARGB) {
	if (((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
	     && info->pixel_stride != 4)
	    || ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_STATE)
		&& (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK))
	    || ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_PREMULTIPLIED_STATE)
		&& !(info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK))) {
	    return SP_FALSE;
	}
	pixel_stride = 4;
	pixel_format = SP_DI_PIXEL_FORMAT_BGRA | SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK;
    } else if (format & PixelFormatAlpha) {
	pixel_stride = 4;
	pixel_format = SP_DI_PIXEL_FORMAT_BGRA;
	format = PixelFormat32bppARGB;
    } else {
	pixel_stride = 3;
	pixel_format = SP_DI_PIXEL_FORMAT_BGR;
	format = PixelFormat24bppRGB;
    }

    spDebug(80, "spGetDIBitmapInfoArch", "format = %x, pixel_format = %lx, pixel_stride = %ld\n", 
	    format, pixel_format, pixel_stride);
    
    if (pixel_format > 0) {
	Gdiplus::BitmapData* gpbitmapData = new Gdiplus::BitmapData;
	Gdiplus::Rect rect(0, 0, gpbitmap->GetWidth(), gpbitmap->GetHeight());
		    
	gpbitmap->LockBits(&rect, Gdiplus::ImageLockModeRead, format, gpbitmapData);
	spDebug(80, "spGetDIBitmapInfoArch", "Stride = %d\n", gpbitmapData->Stride);
	flag = SP_TRUE;
	if (gpbitmapData->Stride > 0) {
	    spDebug(80, "spGetDIBitmapInfoArch", "upside up image, upside_down = %d\n", info->upside_down);
	    if ((*strict_mask & SP_DI_BITMAP_STRICT_UPSIDE_DOWN_STATE)
		&& info->upside_down != SP_FALSE) {
		flag = SP_FALSE;
	    } else {
		info->upside_down = SP_FALSE;
		info->line_stride = (long)gpbitmapData->Stride;
	    }
	} else {
	    spDebug(80, "spGetDIBitmapInfoArch", "upside down image, upside_down = %d\n", info->upside_down);
	    if ((*strict_mask & SP_DI_BITMAP_STRICT_UPSIDE_DOWN_STATE)
		&& info->upside_down == SP_FALSE) {
		flag = SP_FALSE;
	    } else {
		info->upside_down = SP_TRUE;
		info->line_stride = (long)(-gpbitmapData->Stride);
	    }
	}
	gpbitmap->UnlockBits(gpbitmapData);
	
	delete gpbitmapData;

	if (flag) {
	    info->width = gpbitmap->GetWidth();
	    info->height = gpbitmap->GetHeight();
	    spDebug(80, "spGetDIBitmapInfoArch", "width = %ld, height = %ld\n", info->width, info->height);

	    info->pixel_format = pixel_format;
	    info->pixel_stride = pixel_stride;
	    info->bit_size = 8;
	}
    }

    *strict_mask |= SP_DI_BITMAP_STRICT_ALL;

    spDebug(80, "spGetDIBitmapInfoArch", "done: flag = %d\n", flag);
    
    return flag;
}

void *spReadDIBitmapArch(void *instance, spDIBitmapInfoInternal *info, void *bitmap)
{
    Gdiplus::PixelFormat format;
    Gdiplus::Bitmap* gpbitmap;
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;

    if (dibinstance->bmp_with_alpha) {
#if defined(SP_USE_SP_FUNCTION_FOR_BMP_WITH_ALPHA)
        if (dibinstance->fp == NULL) return NULL;

        if (spReadBmpV5Data(&dibinstance->header, info->pixel_format, info->line_stride,
                            info->upside_down, info->data, dibinstance->fp) <= 0) {
            return NULL;
        }

        return bitmap;
#endif
    }
    
    if ((gpbitmap = dibinstance->bitmap) == NULL) return NULL;
	
    spDebug(80, "spReadDIBitmapArch", "width = %ld, height = %ld, pixel_format = %lx, pixel_stride = %ld\n",
	    info->width, info->height, info->pixel_format, info->pixel_stride);

    if (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) {
        if (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK) {
            format = PixelFormat32bppPARGB;
        } else {
            format = PixelFormat32bppARGB;
        }
    } else {
        if (info->pixel_stride == 4) {
            format = PixelFormat32bppRGB;
        } else {
            format = PixelFormat24bppRGB;
        }
    }

    Gdiplus::BitmapData* bitmapData = new Gdiplus::BitmapData;
    Gdiplus::Rect rect(0, 0, gpbitmap->GetWidth(), gpbitmap->GetHeight());
		    
    gpbitmap->LockBits(&rect, Gdiplus::ImageLockModeRead, format, bitmapData);
	    
    spDebug(50, "spReadDIBitmapArch", "bitmapData->Stride = %ld, info->line_stride = %ld\n",
	    (long)bitmapData->Stride, info->line_stride);

    if (bitmapData->Stride < 0) {
        memcpy(info->data, (char *)bitmapData->Scan0 + bitmapData->Stride * (info->height - 1), info->line_stride * info->height);
    } else {
        memcpy(info->data, bitmapData->Scan0, info->line_stride * info->height);
    }

    gpbitmap->UnlockBits(bitmapData);
    
    delete bitmapData;
    
    return bitmap;
}

spBool spWriteDIBitmapArch(void *instance, spDIBitmapInfoInternal *info, void *bitmap)
{
    spBool flag = SP_FALSE;
    INT stride;
    BYTE *ptr;
    Gdiplus::PixelFormat format;
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;

    if (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) {
	if (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK) {
	    format = PixelFormat32bppPARGB;
	} else {
	    format = PixelFormat32bppARGB;
	}
    } else {
	if (info->pixel_stride == 4) {
	    format = PixelFormat32bppRGB;
	} else {
	    format = PixelFormat24bppRGB;
	}
    }

    if (info->upside_down) {
        stride = -info->line_stride;
        ptr = (BYTE *)info->data + info->line_stride * (info->height - 1);
    } else {
        stride = info->line_stride;
        ptr = (BYTE *)info->data;
    }
    
    Gdiplus::Bitmap gpbitmap((INT)info->width, (INT)info->height, stride, format, ptr);
    if (gpbitmap.GetLastStatus() == Gdiplus::Ok) {
	if (gpbitmap.Save((const WCHAR *)dibinstance->wfilename, (const CLSID *)&dibinstance->clsid, NULL) == Gdiplus::Ok) {
	    flag = SP_TRUE;
	}
    }
	
    return flag;
}

